import numpy as np
import scipy.sparse as sp
from scipy.sparse.linalg import spsolve
from scipy.sparse import spmatrix
import copy
import matplotlib.pyplot as plt
import matplotlib.animation as animation

import mesh
from darcy_1D import getmatrhs1D



def time_loop(mesh, K, bcfunc, rhsfun, t0, tf, dt, theta, uinit, lineaire=True):

    t = t0
    ts = [t0]
    u = [uinit]
    uold = uinit
    while(t < tf):
        if (1.1*dt >= tf -t):
            # mieux t = t + min(dt, tf-t)
            dt = tf -t

        bcleft, bcright = bcfunc
        lval = theta*bcleft['value'](t+dt) + (1-theta)*bcleft[ 'value'](t)
        rval = theta*bcright['value'](t+dt) + (1-theta)*bcright[ 'value'](t)
        bcl = {'type': bcleft['type'], 'value': lval}
        bcr = {'type': bcright['type'], 'value': rval}
        bcs = [bcl, bcr]
        if lineaire:
            unew = do_step(mesh, K, bcs, rhsfun(t+dt), dt, theta, uold)
        else:
            unew, critere = picard(mesh, K, bcs, rhsfun(t+dt), dt, theta, uold)
            #print("done ", t, "    critere= " , critere)

        
        u.append(unew)
        t += dt
        ts.append(t)
        uold = unew

    return ts, u

def do_step(mesh, K, bclist, rhsfun, dt, theta, uold):

    A, f = getmatrhs1D(mesh, K, bclist, rhsfun)
    if theta == 0:
        unew = uold + (dt*f -dt*A.dot(uold))/mesh.mes
    else:
        b = 1/theta * (f + mesh.mes*uold/dt - (1-theta)*A.dot(uold))
        B = A.copy()
        D =A.diagonal()
        D += mesh.mes/(theta*dt)
        B.setdiag(D)
        unew = spsolve(B, b)
    return unew

def picard(mesh, Kfun, bclist, rhsfun, dt, theta, uold):
    converged = False
    tol =1e-5

    ukold = uold

    while not(converged):
        # mettre à jour le nouveau K en fonction de ukold
        #puis faire comme au dessus (theta>0)
        # la solution est uknew
        
        # evaluer la convegence, différence relative entre uknew et ukold
        
        ukold = uknew

    return uknew, critere
    

def test_heterogene():
    L = 1
    n = 100
    umesh = mesh.make_uniform_mesh(L, n)
    t0 = 0
    tf = 10000
    dt = 100
    theta = 1
    K = 1e-5*np.ones_like(umesh.centres)
    K[0:n//4] *= 1
    K[n//4:2*n//3] *= 10
    K[2*n//3:n] *= 100
    #K[0:n // 2] = 1e-4
    #K[n // 2:n] = 1e-5

    uinit = np.zeros_like(umesh.centres)
    uinit[0] = 50
    uinit[-1] = 50

    bcleft = {'type': 'Dirichlet', 'value': 100.}
    bcright = {'type': 'Dirichlet', 'value': 100}

    bclist = [bcleft, bcright]
    #   bclist = [bcleftD, bcrightN]
    #   bclist = [bcleftN, bcrightD]
    #   bclist = [bcleftN, bcrightN]

    def rhsfunc(t):
        return np.zeros_like(umesh.centres)

    fig = plt.figure()
    ax = fig.add_subplot(1,1,1)
    line, = ax.plot([], [], 'o-', lw=2)

    ts, u = time_loop(umesh, K, bclist, rhsfunc, t0, tf, dt, theta, uinit)

    def animate(frame):
        ax.clear()
        plt.plot(umesh.centres, u[frame], label='Temp. calculee')
        plt.ylim([0, 100])
        plt.xlim([0, L])
        plt.legend()

    ani = animation.FuncAnimation(fig, animate, frames=len(u), interval=200,repeat=False)
    plt.show()

def test_homogene():

    L = 1
    t0 = 0
    tf = .1
    K0 = 1.

    def initfunc(x):
        return np.sin(np.pi*x/2) +0.5*np.sin(2*np.pi*x)
    
    def rhsfunc(x, t):
        return np.zeros_like(x)

    def solex(x, t):
        return np.exp(-np.pi**2*t/4) * np. sin(np.pi*x/2.) + 0.5 * np.exp(-4*np.pi**2*t) * np. sin(2*np.pi*x)

    bcleft = {'type': 'Dirichlet', 'value': lambda t: solex(0, t)}
    bcright = {'type': 'Dirichlet', 'value': lambda t: solex(L, t)}
    bcfunc = [bcleft, bcright]

    def animate(frame):
        ax.clear()
        plt.plot(umesh.centres, u[frame], '-+', umesh.centres, solex(x, ts[frame]), '-')
        #plt.plot(umesh.centres, np.abs(u[frame] - solex(umesh.centres, ts[frame])), '-')
        plt.ylim([0, 1])
        plt.xlim([0, L])
        plt.xlabel('x')
        plt.ylabel('Temp')
        plt.legend(['Temp. calculee', 'Temp. exacte'])

        
    n = 10
    umesh = mesh.make_uniform_mesh(L, n)
    x = umesh.centres
    dt = 0.01#/4
    theta = 0.5
    K = K0*np.ones_like(x)
    uinit = initfunc(x)

    ts, u = time_loop(umesh, K, bcfunc, lambda t: rhsfunc, t0, tf, dt, theta, uinit)

    h=L/n
    print("Fourier= ", K0*dt/h**2)
    errx_Linf=0
    for i in range(len(u)):
        errxt_Linf = np.max(mesh.normeL2(umesh, u[i]-solex(x, ts[i])))
    print("erreurs: L_inf / L2 ", errxt_Linf)

    fig, ax = plt.subplots()
    ani = animation.FuncAnimation(fig, animate, frames=len(u), interval=40,repeat=False)
    plt.show()



    
if __name__ == '__main__':
    test_homogene()
